# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals

from typing import List, Optional
from matrix.globals import SERVERS, W, SCRIPT_NAME
from matrix.utf import utf8_decode
from matrix.utils import tags_from_line_data
from nio import LocalProtocolError


def add_servers_to_completion(completion):
    for server_name in SERVERS:
        W.hook_completion_list_add(
            completion, server_name, 0, W.WEECHAT_LIST_POS_SORT
        )


@utf8_decode
def matrix_server_command_completion_cb(
    data, completion_item, buffer, completion
):
    buffer_input = W.buffer_get_string(buffer, "input").split()

    args = buffer_input[1:]
    commands = ["add", "delete", "list", "listfull"]

    def complete_commands():
        for command in commands:
            W.hook_completion_list_add(
                completion, command, 0, W.WEECHAT_LIST_POS_SORT
            )

    if len(args) == 1:
        complete_commands()

    elif len(args) == 2:
        if args[1] not in commands:
            complete_commands()
        else:
            if args[1] == "delete" or args[1] == "listfull":
                add_servers_to_completion(completion)

    elif len(args) == 3:
        if args[1] == "delete" or args[1] == "listfull":
            if args[2] not in SERVERS:
                add_servers_to_completion(completion)

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_server_completion_cb(data, completion_item, buffer, completion):
    add_servers_to_completion(completion)
    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_command_completion_cb(data, completion_item, buffer, completion):
    for command in [
        "connect",
        "disconnect",
        "reconnect",
        "server",
        "help",
        "debug",
    ]:
        W.hook_completion_list_add(
            completion, command, 0, W.WEECHAT_LIST_POS_SORT
        )
    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_debug_completion_cb(data, completion_item, buffer, completion):
    for debug_type in ["messaging", "network", "timing"]:
        W.hook_completion_list_add(
            completion, debug_type, 0, W.WEECHAT_LIST_POS_SORT
        )
    return W.WEECHAT_RC_OK


# TODO this should be configurable
REDACTION_COMP_LEN = 50


@utf8_decode
def matrix_message_completion_cb(data, completion_item, buffer, completion):
    max_events = 500

    def redacted_or_not_message(tags):
        # type: (List[str]) -> bool
        if SCRIPT_NAME + "_redacted" in tags:
            return True
        if SCRIPT_NAME + "_message" not in tags:
            return True

        return False

    def event_id_from_tags(tags):
        # type: (List[str]) -> Optional[str]
        for tag in tags:
            if tag.startswith("matrix_id"):
                event_id = tag[10:]
                return event_id

        return None

    for server in SERVERS.values():
        if buffer in server.buffers.values():
            room_buffer = server.find_room_from_ptr(buffer)
            lines = room_buffer.weechat_buffer.lines

            added = 0

            for line in lines:
                tags = line.tags
                if redacted_or_not_message(tags):
                    continue

                event_id = event_id_from_tags(tags)

                if not event_id:
                    continue

                # Make sure we'll be able to reliably detect the end of the
                # quoted snippet
                message_fmt = line.message.replace("\\", "\\\\") \
                                          .replace('"', '\\"')

                if len(message_fmt) > REDACTION_COMP_LEN + 2:
                    message_fmt = message_fmt[:REDACTION_COMP_LEN] + ".."

                item = ('{event_id}|"{message}"').format(
                    event_id=event_id, message=message_fmt
                )

                W.hook_completion_list_add(
                    completion, item, 0, W.WEECHAT_LIST_POS_END
                )
                added += 1

                if added >= max_events:
                    break

            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


def server_from_buffer(buffer):
    for server in SERVERS.values():
        if buffer in server.buffers.values():
            return server
        if buffer == server.server_buffer:
            return server
    return None


@utf8_decode
def matrix_olm_user_completion_cb(data, completion_item, buffer, completion):
    server = server_from_buffer(buffer)

    if not server:
        return W.WEECHAT_RC_OK

    try:
        device_store = server.client.device_store
    except LocalProtocolError:
        return W.WEECHAT_RC_OK

    for user in device_store.users:
        W.hook_completion_list_add(
            completion, user, 0, W.WEECHAT_LIST_POS_SORT
        )

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_olm_device_completion_cb(data, completion_item, buffer, completion):
    server = server_from_buffer(buffer)

    if not server:
        return W.WEECHAT_RC_OK

    try:
        device_store = server.client.device_store
    except LocalProtocolError:
        return W.WEECHAT_RC_OK

    args = W.hook_completion_get_string(completion, "args")

    fields = args.split()

    if len(fields) < 2:
        return W.WEECHAT_RC_OK

    user = fields[-1]

    if user not in device_store.users:
        return W.WEECHAT_RC_OK

    for device in device_store.active_user_devices(user):
        W.hook_completion_list_add(
            completion, device.id, 0, W.WEECHAT_LIST_POS_SORT
        )

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_own_devices_completion_cb(
    data,
    completion_item,
    buffer,
    completion
):
    server = server_from_buffer(buffer)

    if not server:
        return W.WEECHAT_RC_OK

    olm = server.client.olm

    if not olm:
        return W.WEECHAT_RC_OK

    W.hook_completion_list_add(
        completion, olm.device_id, 0, W.WEECHAT_LIST_POS_SORT
    )

    user = olm.user_id

    if user not in olm.device_store.users:
        return W.WEECHAT_RC_OK

    for device in olm.device_store.active_user_devices(user):
        W.hook_completion_list_add(
            completion, device.id, 0, W.WEECHAT_LIST_POS_SORT
        )

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_user_completion_cb(data, completion_item, buffer, completion):
    def add_user(completion, user):
        W.hook_completion_list_add(
            completion, user, 0, W.WEECHAT_LIST_POS_SORT
        )

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            return W.WEECHAT_RC_OK

        room_buffer = server.find_room_from_ptr(buffer)

        if not room_buffer:
            continue

        users = room_buffer.room.users

        users = [user[1:] for user in users]

        for user in users:
            add_user(completion, user)

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_room_completion_cb(data, completion_item, buffer, completion):
    """Completion callback for matrix room names."""
    for server in SERVERS.values():
        for room_buffer in server.room_buffers.values():
            name = room_buffer.weechat_buffer.short_name

            W.hook_completion_list_add(
                completion, name, 0, W.WEECHAT_LIST_POS_SORT
            )

    return W.WEECHAT_RC_OK


def init_completion():
    W.hook_completion(
        "matrix_server_commands",
        "Matrix server completion",
        "matrix_server_command_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_servers",
        "Matrix server completion",
        "matrix_server_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_commands",
        "Matrix command completion",
        "matrix_command_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_messages",
        "Matrix message completion",
        "matrix_message_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_debug_types",
        "Matrix debugging type completion",
        "matrix_debug_completion_cb",
        "",
    )

    W.hook_completion(
        "olm_user_ids",
        "Matrix olm user id completion",
        "matrix_olm_user_completion_cb",
        "",
    )

    W.hook_completion(
        "olm_devices",
        "Matrix olm device id completion",
        "matrix_olm_device_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_users",
        "Matrix user id completion",
        "matrix_user_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_own_devices",
        "Matrix own devices completion",
        "matrix_own_devices_completion_cb",
        "",
    )

    W.hook_completion(
        "matrix_rooms",
        "Matrix room name completion",
        "matrix_room_completion_cb",
        "",
    )



# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
# Copyright © 2018, 2019 Denis Kasak <dkasak@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals, division

import time
from typing import Any, Dict, List

from .globals import W

if False:
    from .server import MatrixServer


def key_from_value(dictionary, value):
    # type: (Dict[str, Any], Any) -> str
    return list(dictionary.keys())[list(dictionary.values()).index(value)]


def server_buffer_prnt(server, string):
    # type: (MatrixServer, str) -> None
    assert server.server_buffer
    buffer = server.server_buffer
    now = int(time.time())
    W.prnt_date_tags(buffer, now, "", string)


def tags_from_line_data(line_data):
    # type: (str) -> List[str]
    tags_count = W.hdata_get_var_array_size(
        W.hdata_get("line_data"), line_data, "tags_array"
    )

    tags = [
        W.hdata_string(
            W.hdata_get("line_data"), line_data, "%d|tags_array" % i
        )
        for i in range(tags_count)
    ]

    return tags


def create_server_buffer(server):
    # type: (MatrixServer) -> None
    buffer_name = "server.{}".format(server.name)
    server.server_buffer = W.buffer_new(
        buffer_name, "server_buffer_cb", server.name, "", ""
    )

    server_buffer_set_title(server)
    W.buffer_set(server.server_buffer, "short_name", server.name)
    W.buffer_set(server.server_buffer, "localvar_set_type", "server")
    W.buffer_set(
        server.server_buffer, "localvar_set_nick", server.config.username
    )
    W.buffer_set(server.server_buffer, "localvar_set_server", server.name)
    W.buffer_set(server.server_buffer, "localvar_set_channel", server.name)

    server.buffer_merge()


def server_buffer_set_title(server):
    # type: (MatrixServer) -> None
    if server.numeric_address:
        ip_string = " ({address})".format(address=server.numeric_address)
    else:
        ip_string = ""

    title = ("Matrix: {address}:{port}{ip}").format(
        address=server.address, port=server.config.port, ip=ip_string
    )

    W.buffer_set(server.server_buffer, "title", title)


def server_ts_to_weechat(timestamp):
    # type: (float) -> int
    date = int(timestamp / 1000)
    return date


def strip_matrix_server(string):
    # type: (str) -> str
    return string.rsplit(":", 1)[0]


def shorten_sender(sender):
    # type: (str) -> str
    return strip_matrix_server(sender)[1:]


def string_strikethrough(string):
    return "".join(["{}\u0336".format(c) for c in string])


def string_color_and_reset(string, color):
    """Color string with color, then reset all attributes."""

    lines = string.split('\n')
    lines = ("{}{}{}".format(W.color(color), line, W.color("reset"))
             for line in lines)
    return "\n".join(lines)


def string_color(string, color):
    """Color string with color, then reset the color attribute."""

    lines = string.split('\n')
    lines = ("{}{}{}".format(W.color(color), line, W.color("resetcolor"))
             for line in lines)
    return "\n".join(lines)


def color_pair(color_fg, color_bg):
    """Make a color pair from a pair of colors."""

    if color_bg:
        return "{},{}".format(color_fg, color_bg)
    else:
        return color_fg


def text_block(text, margin=0):
    """
    Pad block of text with whitespace to form a regular block, optionally
    adding a margin.
    """

    # add vertical margin
    vertical_margin = margin // 2
    text = "{}{}{}".format(
        "\n" * vertical_margin,
        text,
        "\n" * vertical_margin
    )

    lines = text.split("\n")
    longest_len = max(len(l) for l in lines) + margin

    # pad block and add horizontal margin
    text = "\n".join(
        "{pre}{line}{post}".format(
            pre=" " * margin,
            line=l,
            post=" " * (longest_len - len(l)))
        for l in lines)

    return text


def colored_text_block(text, margin=0, color_pair=""):
    """ Like text_block, but also colors it."""
    return string_color_and_reset(text_block(text, margin=margin), color_pair)

def parse_redact_args(args):
    args = args.strip()

    had_example_text = False

    try:
        event_id, rest = args.split("|", 1)
        had_example_text = True
    except ValueError:
        try:
            event_id, rest = args.split(" ", 1)
        except ValueError:
            event_id, rest = (args, "")

    if had_example_text:
        rest = rest.lstrip()
        reason = None  # until it has been correctly determined
        if rest[0] == '"':
            escaped = False
            for i in range(1, len(rest)):
                if escaped:
                    escaped = False
                elif rest[i] == "\\":
                    escaped = True
                elif rest[i] == '"':
                    reason = rest[i+1:]
                    break
    else:
        reason = rest

    event_id = event_id.strip()
    if reason:
        reason = reason.strip()
    # The reason might be an empty string, set it to None if so
    else:
        reason = None

    return event_id, reason



# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
# Copyright © 2018, 2019 Denis Kasak <dkasak@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""weechat-matrix Configuration module.

This module contains abstractions on top of weechats configuration files and
the main script configuration class.

To add configuration options refer to MatrixConfig.
Server specific configuration options are handled in server.py
"""

from builtins import super
from collections import namedtuple
from enum import IntEnum, Enum, unique

import logbook

import nio
from matrix.globals import SCRIPT_NAME, SERVERS, W
from matrix.utf import utf8_decode

from . import globals as G


@unique
class RedactType(Enum):
    STRIKETHROUGH = 0
    NOTICE = 1
    DELETE = 2


@unique
class ServerBufferType(Enum):
    MERGE_CORE = 0
    MERGE = 1
    INDEPENDENT = 2


@unique
class NewChannelPosition(IntEnum):
    NONE = 0
    NEXT = 1
    NEAR_SERVER = 2


nio.logger_group.level = logbook.ERROR


class Option(
    namedtuple(
        "Option",
        [
            "name",
            "type",
            "string_values",
            "min",
            "max",
            "value",
            "description",
            "cast_func",
            "change_callback",
        ],
    )
):
    """A class representing a new configuration option.

    An option object is consumed by the ConfigSection class adding
    configuration options to weechat.
    """

    __slots__ = ()

    def __new__(
        cls,
        name,
        type,
        string_values,
        min,
        max,
        value,
        description,
        cast=None,
        change_callback=None,
    ):
        """
        Parameters:
            name (str): Name of the configuration option
            type (str): Type of the configuration option, can be one of the
                supported weechat types: string, boolean, integer, color
            string_values: (str): A list of string values that the option can
            accept seprated by |
            min (int): Minimal value of the option, only used if the type of
                the option is integer
            max (int): Maximal value of the option, only used if the type of
                the option is integer
            description (str): Description of the configuration option
            cast (callable): A callable function taking a single value and
                returning a modified value. Useful to turn the configuration
                option into an enum while reading it.
            change_callback(callable): A function that will be called
                by weechat every time the configuration option is changed.
        """

        return super().__new__(
            cls,
            name,
            type,
            string_values,
            min,
            max,
            value,
            description,
            cast,
            change_callback,
        )


@utf8_decode
def matrix_config_reload_cb(data, config_file):
    return W.WEECHAT_RC_OK


def change_log_level(category, level):
    """Change the log level of the underlying nio lib

    Called every time the user changes the log level or log category
    configuration option."""

    if category == "all":
        nio.logger_group.level = level
    elif category == "http":
        nio.http.logger.level = level
    elif category == "client":
        nio.client.logger.level = level
    elif category == "events":
        nio.events.logger.level = level
    elif category == "responses":
        nio.responses.logger.level = level
    elif category == "encryption":
        nio.crypto.logger.level = level


@utf8_decode
def config_server_buffer_cb(data, option):
    """Callback for the look.server_buffer option.
    Is called when the option is changed and merges/splits the server
    buffer"""

    for server in SERVERS.values():
        server.buffer_merge()
    return 1


@utf8_decode
def config_log_level_cb(data, option):
    """Callback for the network.debug_level option."""
    change_log_level(
        G.CONFIG.network.debug_category, G.CONFIG.network.debug_level
    )
    return 1


@utf8_decode
def config_log_category_cb(data, option):
    """Callback for the network.debug_category option."""
    change_log_level(G.CONFIG.debug_category, logbook.ERROR)
    G.CONFIG.debug_category = G.CONFIG.network.debug_category
    change_log_level(
        G.CONFIG.network.debug_category, G.CONFIG.network.debug_level
    )
    return 1


@utf8_decode
def config_pgup_cb(data, option):
    """Callback for the network.fetch_backlog_on_pgup option.
    Enables or disables the hook that is run when /window page_up is called"""
    if G.CONFIG.network.fetch_backlog_on_pgup:
        if not G.CONFIG.page_up_hook:
            G.CONFIG.page_up_hook = W.hook_command_run(
                "/window page_up", "matrix_command_pgup_cb", ""
            )
    else:
        if G.CONFIG.page_up_hook:
            W.unhook(G.CONFIG.page_up_hook)
            G.CONFIG.page_up_hook = None

    return 1


def level_to_logbook(value):
    if value == 0:
        return logbook.ERROR
    if value == 1:
        return logbook.WARNING
    if value == 2:
        return logbook.INFO
    if value == 3:
        return logbook.DEBUG

    return logbook.ERROR


def logbook_category(value):
    if value == 0:
        return "all"
    if value == 1:
        return "http"
    if value == 2:
        return "client"
    if value == 3:
        return "events"
    if value == 4:
        return "responses"
    if value == 5:
        return "encryption"

    return "all"


def parse_nick_prefix_colors(value):
    """Parses the nick prefix color setting string
    ("admin=COLOR1;mod=COLOR2;power=COLOR3") into a prefix -> color dict."""

    def key_to_prefix(key):
        if key == "admin":
            return "&"
        elif key == "mod":
            return "@"
        elif key == "power":
            return "+"
        else:
            return ""

    prefix_colors = {
        "&": "lightgreen",
        "@": "lightgreen",
        "+": "yellow",
    }

    for setting in value.split(";"):
        # skip malformed settings
        if "=" not in setting:
            continue

        key, color = setting.split("=")
        prefix = key_to_prefix(key)

        if prefix:
            prefix_colors[prefix] = color

    return prefix_colors


def eval_cast(string):
    """A function that passes a string to weechat which evaluates it using its
    expression evaluation syntax.
    Can only be used with strings, useful for passwords or options that contain
    a formatted string to e.g. add colors.
    More info here:
        https://weechat.org/files/doc/stable/weechat_plugin_api.en.html#_string_eval_expression"""

    return W.string_eval_expression(string, {}, {}, {})


class WeechatConfig(object):
    """A class representing a weechat configuration file
    Wraps weechats configuration creation functionality"""

    def __init__(self, sections):
        """Create a new weechat configuration file, expects the global
        SCRIPT_NAME to be defined and a reload callback

        Parameters:
            sections (List[Tuple[str, List[Option]]]): List of config sections
                that will be created for the configuration file.
        """
        self._ptr = W.config_new(
            SCRIPT_NAME, SCRIPT_NAME + "_config_reload_cb", ""
        )

        for section in sections:
            name, options = section
            section_class = ConfigSection.build(name, options)
            setattr(self, name, section_class(name, self._ptr, options))

    def free(self):
        """Free all the config sections and their options as well as the
        configuration file. Should be called when the script is unloaded."""
        for section in [
            getattr(self, a)
            for a in dir(self)
            if isinstance(getattr(self, a), ConfigSection)
        ]:
            section.free()

        W.config_free(self._ptr)

    def read(self):
        """Read the config file"""
        return_code = W.config_read(self._ptr)
        if return_code == W.WEECHAT_CONFIG_READ_OK:
            return True
        if return_code == W.WEECHAT_CONFIG_READ_MEMORY_ERROR:
            return False
        if return_code == W.WEECHAT_CONFIG_READ_FILE_NOT_FOUND:
            return True
        return False


class ConfigSection(object):
    """A class representing a weechat config section.
    Should not be used on its own, the WeechatConfig class uses this to build
    config sections."""
    @classmethod
    def build(cls, name, options):
        def constructor(self, name, config_ptr, options):
            self._ptr = W.config_new_section(
                config_ptr, name, 0, 0, "", "", "", "", "", "", "", "", "", ""
            )
            self._config_ptr = config_ptr
            self._option_ptrs = {}

            for option in options:
                self._add_option(option)

        attributes = {
            option.name: cls.option_property(
                option.name, option.type, cast_func=option.cast_func
            )
            for option in options
        }
        attributes["__init__"] = constructor

        section_class = type(name.title() + "Section", (cls,), attributes)
        return section_class

    def free(self):
        W.config_section_free_options(self._ptr)
        W.config_section_free(self._ptr)

    def _add_option(self, option):
        cb = option.change_callback.__name__ if option.change_callback else ""
        option_ptr = W.config_new_option(
            self._config_ptr,
            self._ptr,
            option.name,
            option.type,
            option.description,
            option.string_values,
            option.min,
            option.max,
            option.value,
            option.value,
            0,
            "",
            "",
            cb,
            "",
            "",
            "",
        )

        self._option_ptrs[option.name] = option_ptr

    @staticmethod
    def option_property(name, option_type, evaluate=False, cast_func=None):
        """Create a property for this class that makes the reading of config
        option values pythonic. The option will be available as a property with
        the name of the option.
        If a cast function was defined for the option the property will pass
        the option value to the cast function and return its result."""

        def bool_getter(self):
            return bool(W.config_boolean(self._option_ptrs[name]))

        def str_getter(self):
            if cast_func:
                return cast_func(W.config_string(self._option_ptrs[name]))
            return W.config_string(self._option_ptrs[name])

        def str_evaluate_getter(self):
            return W.string_eval_expression(
                W.config_string(self._option_ptrs[name]), {}, {}, {}
            )

        def int_getter(self):
            if cast_func:
                return cast_func(W.config_integer(self._option_ptrs[name]))
            return W.config_integer(self._option_ptrs[name])

        if option_type in ("string", "color"):
            if evaluate:
                return property(str_evaluate_getter)
            return property(str_getter)
        if option_type == "boolean":
            return property(bool_getter)
        if option_type == "integer":
            return property(int_getter)


class MatrixConfig(WeechatConfig):
    """Main matrix configuration file.
    This class defines all the global matrix configuration options.
    New global options should be added to the constructor of this class under
    the appropriate section.

    There are three main sections defined:
        Look: This section is for options that change the way matrix messages
            are shown or the way the buffers are shown.
        Color: This section should mainly be for color options, options that
            change color schemes or themes should go to the look section.
        Network: This section is for options that change the way the script
            behaves, e.g. the way it communicates with the server, it handles
            responses or any other behavioural change that doesn't fit in the
            previous sections.

    There is a special section called server defined which contains per server
    configuration options. Server options aren't defined here, they need to be
    added in server.py
    """

    def __init__(self):
        self.debug_buffer = ""
        self.upload_buffer = ""
        self.debug_category = "all"
        self.page_up_hook = None
        self.human_buffer_names = None

        look_options = [
            Option(
                "redactions",
                "integer",
                "strikethrough|notice|delete",
                0,
                0,
                "strikethrough",
                (
                    "Only notice redactions, strike through or delete "
                    "redacted messages"
                ),
                RedactType,
            ),
            Option(
                "server_buffer",
                "integer",
                "merge_with_core|merge_without_core|independent",
                0,
                0,
                "merge_with_core",
                "Merge server buffers",
                ServerBufferType,
                config_server_buffer_cb,
            ),
            Option(
                "new_channel_position",
                "integer",
                "none|next|near_server",
                min(NewChannelPosition),
                max(NewChannelPosition),
                "none",
                "force position of new channel in list of buffers "
                "(none = default position (should be last buffer), "
                "next = current buffer + 1, near_server = after last "
                "channel/pv of server)",
                NewChannelPosition,
            ),
            Option(
                "max_typing_notice_item_length",
                "integer",
                "",
                10,
                1000,
                "50",
                ("Limit the length of the typing notice bar item."),
            ),
            Option(
                "bar_item_typing_notice_prefix",
                "string",
                "",
                0,
                0,
                "Typing: ",
                ("Prefix for the typing notice bar item."),
            ),
            Option(
                "encryption_warning_sign",
                "string",
                "",
                0,
                0,
                "⚠️ ",
                ("A sign that is used to signal trust issues in encrypted "
                 "rooms (note: content is evaluated, see /help eval)"),
                eval_cast,
            ),
            Option(
                "busy_sign",
                "string",
                "",
                0,
                0,
                "⏳",
                ("A sign that is used to signal that the client is busy e.g. "
                 "when the room backlog is fetching"
                 " (note: content is evaluated, see /help eval)"),
                eval_cast,
            ),
            Option(
                "encrypted_room_sign",
                "string",
                "",
                0,
                0,
                "🔐",
                ("A sign that is used to show that the current room is "
                 "encrypted "
                 "(note: content is evaluated, see /help eval)"),
                eval_cast,
            ),
            Option(
                "disconnect_sign",
                "string",
                "",
                0,
                0,
                "❌",
                ("A sign that is used to show that the server is disconnected "
                 "(note: content is evaluated, see /help eval)"),
                eval_cast,
            ),
            Option(
                "pygments_style",
                "string",
                "",
                0,
                0,
                "native",
                "Pygments style to use for highlighting source code blocks",
            ),
            Option(
                "code_blocks",
                "boolean",
                "",
                0,
                0,
                "on",
                ("Display preformatted code blocks as rectangular areas by "
                 "padding them with whitespace up to the length of the longest"
                 " line (with optional margin)"),
            ),
            Option(
                "code_block_margin",
                "integer",
                "",
                0,
                100,
                "2",
                ("Number of spaces to add as a margin around around a code "
                 "block"),
            ),
            Option(
                "quote_wrap",
                "integer",
                "",
                -1,
                1000,
                "67",
                ("After how many characters to soft-wrap lines in a quote "
                 "block (reply message). Set to -1 to disable soft-wrapping."),
            ),
            Option(
                "human_buffer_names",
                "boolean",
                "",
                0,
                0,
                "off",
                ("If turned on the buffer name will consist of the server "
                 "name and the room name instead of the Matrix room ID. Note, "
                 "this requires a change to the logger.file.mask setting "
                 "since conflicts can happen otherwise "
                 "(requires a script reload)."),
            ),
            Option(
                "markdown_input",
                "boolean",
                "",
                0,
                0,
                "on",
                ("If turned on, markdown usage in messages will be converted "
                 "to actual markup (**bold**, *italic*, _italic_, `code`)."),
            ),
        ]

        network_options = [
            Option(
                "max_initial_sync_events",
                "integer",
                "",
                1,
                10000,
                "30",
                ("How many events to fetch during the initial sync"),
            ),
            Option(
                "max_backlog_sync_events",
                "integer",
                "",
                1,
                100,
                "10",
                ("How many events to fetch during backlog fetching"),
            ),
            Option(
                "fetch_backlog_on_pgup",
                "boolean",
                "",
                0,
                0,
                "on",
                ("Fetch messages in the backlog on a window page up event"),
                None,
                config_pgup_cb,
            ),
            Option(
                "debug_level",
                "integer",
                "error|warn|info|debug",
                0,
                0,
                "error",
                "Enable network protocol debugging.",
                level_to_logbook,
                config_log_level_cb,
            ),
            Option(
                "debug_category",
                "integer",
                "all|http|client|events|responses|encryption",
                0,
                0,
                "all",
                "Debugging category",
                logbook_category,
                config_log_category_cb,
            ),
            Option(
                "debug_buffer",
                "boolean",
                "",
                0,
                0,
                "off",
                ("Use a separate buffer for debug logs."),
            ),
            Option(
                "lazy_load_room_users",
                "boolean",
                "",
                0,
                0,
                "off",
                ("If on, room users won't be loaded in the background "
                 "proactively, they will be loaded when the user switches to "
                 "the room buffer. This only affects non-encrypted rooms."),
            ),
            Option(
                "max_nicklist_users",
                "integer",
                "",
                100,
                20000,
                "5000",
                ("Limit the number of users that are added to the nicklist. "
                 "Active users and users with a higher power level are always."
                 " Inactive users will be removed from the nicklist after a "
                 "day of inactivity."),
            ),
            Option(
                "lag_reconnect",
                "integer",
                "",
                5,
                604800,
                "90",
                ("Reconnect to the server if the lag is greater than this "
                 "value (in seconds)"),
            ),
            Option(
                "autoreconnect_delay_growing",
                "integer",
                "",
                1,
                100,
                "2",
                ("growing factor for autoreconnect delay to server "
                 "(1 = always same delay, 2 = delay*2 for each retry, etc.)"),
            ),
            Option(
                "autoreconnect_delay_max",
                "integer",
                "",
                0,
                604800,
                "600",
                ("maximum autoreconnect delay to server "
                 "(in seconds, 0 = no maximum)"),
            ),
            Option(
                "print_unconfirmed_messages",
                "boolean",
                "",
                0,
                0,
                "on",
                ("If off, messages are only printed after the server confirms "
                 "their receival. If on, messages are immediately printed but "
                 "colored differently until receival is confirmed."),
            ),
            Option(
                "lag_min_show",
                "integer",
                "",
                1,
                604800,
                "500",
                ("minimum lag to show (in milliseconds)"),
            ),
            Option(
                "typing_notice_conditions",
                "string",
                "",
                0,
                0,
                "${typing_enabled}",
                ("conditions to send typing notifications (note: content is "
                 "evaluated, see /help eval); besides the buffer and window "
                 "variables the typing_enabled variable is also expanded; "
                 "the typing_enabled variable can be manipulated with the "
                 "/room command, see /help room"),
            ),
            Option(
                "read_markers_conditions",
                "string",
                "",
                0,
                0,
                "${markers_enabled}",
                ("conditions to send read markers (note: content is "
                 "evaluated, see /help eval); besides the buffer and window "
                 "variables the markers_enabled variable is also expanded; "
                 "the markers_enabled variable can be manipulated with the "
                 "/room command, see /help room"),
            ),
            Option(
                "resending_ignores_devices",
                "boolean",
                "",
                0,
                0,
                "on",
                ("If on resending the same message to a room that contains "
                 "unverified devices will mark the devices as ignored and "
                 "continue sending the message. If off resending the message "
                 "will again fail and devices need to be marked as verified "
                 "one by one or the /send-anyways command needs to be used to "
                 "ignore them."),
            ),
        ]

        color_options = [
            Option(
                "quote_fg",
                "color",
                "",
                0,
                0,
                "lightgreen",
                "Foreground color for matrix style blockquotes",
            ),
            Option(
                "quote_bg",
                "color",
                "",
                0,
                0,
                "default",
                "Background counterpart of quote_fg",
            ),
            Option(
                "error_message_fg",
                "color",
                "",
                0,
                0,
                "darkgray",
                ("Foreground color for error messages that appear inside a "
                 "room buffer (e.g. when a message errors out when sending or "
                 "when a message is redacted)"),
            ),
            Option(
                "error_message_bg",
                "color",
                "",
                0,
                0,
                "default",
                "Background counterpart of error_message_fg.",
            ),
            Option(
                "unconfirmed_message_fg",
                "color",
                "",
                0,
                0,
                "darkgray",
                ("Foreground color for messages that are printed out but the "
                 "server hasn't confirmed the that he received them."),
            ),
            Option(
                "unconfirmed_message_bg",
                "color",
                "",
                0,
                0,
                "default",
                "Background counterpart of unconfirmed_message_fg."
            ),
            Option(
                "untagged_code_fg",
                "color",
                "",
                0,
                0,
                "blue",
                ("Foreground color for code without a language specifier. "
                 "Also used for `inline code`."),
            ),
            Option(
                "untagged_code_bg",
                "color",
                "",
                0,
                0,
                "default",
                "Background counterpart of untagged_code_fg",
            ),
            Option(
                "nick_prefixes",
                "string",
                "",
                0,
                0,
                "admin=lightgreen;mod=lightgreen;power=yellow",
                ('Colors for nick prefixes indicating power level. '
                 'Format is "admin:color1;mod:color2;power:color3", '
                 'where "admin" stands for admins (power level = 100), '
                 '"mod" stands for moderators (power level >= 50) and '
                 '"power" for any other power user (power level > 0). '
                 'Requires restart to apply changes.'),
                parse_nick_prefix_colors,
            ),
        ]

        sections = [
            ("network", network_options),
            ("look", look_options),
            ("color", color_options),
        ]

        super().__init__(sections)

        # The server section is essentially a section with subsections and no
        # options, handle that case independently.
        W.config_new_section(
            self._ptr,
            "server",
            0,
            0,
            "matrix_config_server_read_cb",
            "",
            "matrix_config_server_write_cb",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
        )

    def read(self):
        super().read()
        self.human_buffer_names = self.look.human_buffer_names

    def free(self):
        section_ptr = W.config_search_section(self._ptr, "server")
        W.config_section_free(section_ptr)
        super().free()



# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals

import sys
from typing import Any, Dict, Optional
from logbook import Logger
from collections import OrderedDict

from .utf import WeechatWrapper

if False:
    from .server import MatrixServer
    from .config import MatrixConfig
    from .uploads import Upload


try:
    import weechat

    W = weechat if sys.hexversion >= 0x3000000 else WeechatWrapper(weechat)
except ImportError:
    import matrix._weechat as weechat  # type: ignore

    W = weechat

SERVERS = dict()  # type: Dict[str, MatrixServer]
CONFIG = None  # type: Any
ENCRYPTION = True  # type: bool
SCRIPT_NAME = "matrix"  # type: str
BUFFER_NAME_PREFIX = "{}.".format(SCRIPT_NAME)  # type: str
TYPING_NOTICE_TIMEOUT = 4000  # 4 seconds typing notice lifetime
LOGGER = Logger("weechat-matrix")
UPLOADS = OrderedDict()  # type: Dict[str, Upload]



# -*- coding: utf-8 -*-

# Weechat Matrix Protocol Script
# Copyright © 2018 Damir Jelić <poljar@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals

import os

# See if there is a `venv` directory next to our script, and use that if
# present. This first resolves symlinks, so this also works when we are
# loaded through a symlink (e.g. from autoload).
# See https://virtualenv.pypa.io/en/latest/userguide/#using-virtualenv-without-bin-python
# This does not support pyvenv or the python3 venv module, which do not
# create an activate_this.py: https://stackoverflow.com/questions/27462582
activate_this = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'venv', 'bin', 'activate_this.py')
if os.path.exists(activate_this):
    exec(open(activate_this).read(), {'__file__': activate_this})

import socket
import ssl
import textwrap
# pylint: disable=redefined-builtin
from builtins import str
from itertools import chain
# pylint: disable=unused-import
from typing import Any, AnyStr, Deque, Dict, List, Optional, Set, Text, Tuple

import logbook
import json
import OpenSSL.crypto as crypto
from future.utils import bytes_to_native_str as n
from logbook import Logger, StreamHandler

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError  # type: ignore


from nio import RemoteProtocolError, RemoteTransportError, TransportType

from matrix import globals as G
from matrix.bar_items import (
    init_bar_items,
    matrix_bar_item_buffer_modes,
    matrix_bar_item_lag,
    matrix_bar_item_name,
    matrix_bar_item_plugin,
    matrix_bar_nicklist_count,
    matrix_bar_typing_notices_cb
)
from matrix.buffer import room_buffer_close_cb, room_buffer_input_cb
# Weechat searches for the registered callbacks in the scope of the main script
# file, import the callbacks here so weechat can find them.
from matrix.commands import (hook_commands, hook_key_bindings, hook_page_up,
                             matrix_command_buf_clear_cb, matrix_command_cb,
                             matrix_command_pgup_cb, matrix_invite_command_cb,
                             matrix_join_command_cb, matrix_kick_command_cb,
                             matrix_me_command_cb, matrix_part_command_cb,
                             matrix_redact_command_cb, matrix_topic_command_cb,
                             matrix_olm_command_cb, matrix_devices_command_cb,
                             matrix_room_command_cb, matrix_uploads_command_cb,
                             matrix_upload_command_cb, matrix_send_anyways_cb,
                             matrix_reply_command_cb,
                             matrix_cursor_reply_signal_cb)
from matrix.completion import (init_completion, matrix_command_completion_cb,
                               matrix_debug_completion_cb,
                               matrix_message_completion_cb,
                               matrix_olm_device_completion_cb,
                               matrix_olm_user_completion_cb,
                               matrix_server_command_completion_cb,
                               matrix_server_completion_cb,
                               matrix_user_completion_cb,
                               matrix_own_devices_completion_cb,
                               matrix_room_completion_cb)
from matrix.config import (MatrixConfig, config_log_category_cb,
                           config_log_level_cb, config_server_buffer_cb,
                           matrix_config_reload_cb, config_pgup_cb)
from matrix.globals import SCRIPT_NAME, SERVERS, W
from matrix.server import (MatrixServer, create_default_server,
                           matrix_config_server_change_cb,
                           matrix_config_server_read_cb,
                           matrix_config_server_write_cb, matrix_timer_cb,
                           send_cb, matrix_load_users_cb)
from matrix.utf import utf8_decode
from matrix.utils import server_buffer_prnt, server_buffer_set_title

from matrix.uploads import UploadsBuffer, upload_cb

try:
    from urllib.parse import urlunparse
except ImportError:
    from urlparse import urlunparse

# yapf: disable
WEECHAT_SCRIPT_NAME = SCRIPT_NAME
WEECHAT_SCRIPT_DESCRIPTION = "matrix chat plugin"              # type: str
WEECHAT_SCRIPT_AUTHOR = "Damir Jelić <poljar@termina.org.uk>"  # type: str
WEECHAT_SCRIPT_VERSION = "0.3.0"                               # type: str
WEECHAT_SCRIPT_LICENSE = "ISC"                                 # type: str
# yapf: enable


logger = Logger("matrix-cli")


def print_certificate_info(buff, sock, cert):
    cert_pem = ssl.DER_cert_to_PEM_cert(sock.getpeercert(True))

    x509 = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem)

    public_key = x509.get_pubkey()

    key_type = ("RSA" if public_key.type() == crypto.TYPE_RSA else "DSA")
    key_size = str(public_key.bits())
    sha256_fingerprint = x509.digest(n(b"SHA256"))
    sha1_fingerprint = x509.digest(n(b"SHA1"))
    signature_algorithm = x509.get_signature_algorithm()

    key_info = ("key info: {key_type} key {bits} bits, signed using "
                "{algo}").format(
                    key_type=key_type, bits=key_size,
                    algo=n(signature_algorithm))

    validity_info = ("        Begins on:  {before}\n"
                     "        Expires on: {after}").format(
                         before=cert["notBefore"], after=cert["notAfter"])

    rdns = chain(*cert["subject"])
    subject = ", ".join(["{}={}".format(name, value) for name, value in rdns])

    rdns = chain(*cert["issuer"])
    issuer = ", ".join(["{}={}".format(name, value) for name, value in rdns])

    subject = "subject: {sub}, serial number {serial}".format(
        sub=subject, serial=cert["serialNumber"])

    issuer = "issuer: {issuer}".format(issuer=issuer)

    fingerprints = ("        SHA1:   {}\n"
                    "        SHA256: {}").format(n(sha1_fingerprint),
                                                 n(sha256_fingerprint))

    wrapper = textwrap.TextWrapper(
        initial_indent="    - ", subsequent_indent="        ")

    message = ("{prefix}matrix: received certificate\n"
               " - certificate info:\n"
               "{subject}\n"
               "{issuer}\n"
               "{key_info}\n"
               "    - period of validity:\n{validity_info}\n"
               "    - fingerprints:\n{fingerprints}").format(
                   prefix=W.prefix("network"),
                   subject=wrapper.fill(subject),
                   issuer=wrapper.fill(issuer),
                   key_info=wrapper.fill(key_info),
                   validity_info=validity_info,
                   fingerprints=fingerprints)

    W.prnt(buff, message)


def wrap_socket(server, file_descriptor):
    # type: (MatrixServer, int) -> None
    sock = None  # type: socket.socket

    temp_socket = socket.fromfd(file_descriptor, socket.AF_INET,
                                socket.SOCK_STREAM)

    # fromfd() duplicates the file descriptor, we can close the one we got from
    # weechat now since we use the one from our socket when calling hook_fd()
    os.close(file_descriptor)

    # For python 2.7 wrap_socket() doesn't work with sockets created from an
    # file descriptor because fromfd() doesn't return a wrapped socket, the bug
    # was fixed for python 3, more info: https://bugs.python.org/issue13942
    # pylint: disable=protected-access,unidiomatic-typecheck
    if type(temp_socket) == socket._socket.socket:
        # pylint: disable=no-member
        sock = socket._socketobject(_sock=temp_socket)
    else:
        sock = temp_socket

    # fromfd() duplicates the file descriptor but doesn't retain it's blocking
    # non-blocking attribute, so mark the socket as non-blocking even though
    # weechat already did that for us
    sock.setblocking(False)

    message = "{prefix}matrix: Doing SSL handshake...".format(
        prefix=W.prefix("network"))
    W.prnt(server.server_buffer, message)

    ssl_socket = server.ssl_context.wrap_socket(
        sock, do_handshake_on_connect=False,
        server_hostname=server.address)  # type: ssl.SSLSocket

    server.socket = ssl_socket

    try_ssl_handshake(server)


@utf8_decode
def ssl_fd_cb(server_name, file_descriptor):
    server = SERVERS[server_name]

    if server.ssl_hook:
        W.unhook(server.ssl_hook)
        server.ssl_hook = None

    try_ssl_handshake(server)

    return W.WEECHAT_RC_OK


def try_ssl_handshake(server):
    sock = server.socket

    while True:
        try:
            sock.do_handshake()

            cipher = sock.cipher()
            cipher_message = ("{prefix}matrix: Connected using {tls}, and "
                              "{bit} bit {cipher} cipher suite.").format(
                                  prefix=W.prefix("network"),
                                  tls=cipher[1],
                                  bit=cipher[2],
                                  cipher=cipher[0])
            W.prnt(server.server_buffer, cipher_message)

            cert = sock.getpeercert()
            if cert:
                print_certificate_info(server.server_buffer, sock, cert)

            finalize_connection(server)

            return True

        except ssl.SSLWantReadError:
            hook = W.hook_fd(server.socket.fileno(), 1, 0, 0, "ssl_fd_cb",
                             server.name)
            server.ssl_hook = hook

            return False

        except ssl.SSLWantWriteError:
            hook = W.hook_fd(server.socket.fileno(), 0, 1, 0, "ssl_fd_cb",
                             server.name)
            server.ssl_hook = hook

            return False

        except (ssl.SSLError, ssl.CertificateError, socket.error) as error:
            try:
                str_error = error.reason if error.reason else "Unknown error"
            except AttributeError:
                str_error = str(error)

            message = ("{prefix}Error while doing SSL handshake"
                       ": {error}").format(
                           prefix=W.prefix("network"), error=str_error)

            server_buffer_prnt(server, message)

            server_buffer_prnt(
                server, ("{prefix}matrix: disconnecting from server..."
                         ).format(prefix=W.prefix("network")))

            server.disconnect()
            return False


@utf8_decode
def receive_cb(server_name, file_descriptor):
    server = SERVERS[server_name]

    while True:
        try:
            data = server.socket.recv(4096)
        except ssl.SSLWantReadError:
            break
        except socket.error as error:
            errno = "error" + str(error.errno) + " " if error.errno else ""
            str_error = error.strerror if error.strerror else "Unknown error"
            str_error = errno + str_error

            message = ("{prefix}Error while reading from "
                       "socket: {error}").format(
                           prefix=W.prefix("network"), error=str_error)

            server_buffer_prnt(server, message)

            server_buffer_prnt(
                server, ("{prefix}matrix: disconnecting from server..."
                         ).format(prefix=W.prefix("network")))

            server.disconnect()

            return W.WEECHAT_RC_OK

        if not data:
            server_buffer_prnt(
                server,
                "{prefix}matrix: Error while reading from socket".format(
                    prefix=W.prefix("network")))
            server_buffer_prnt(
                server, ("{prefix}matrix: disconnecting from server..."
                         ).format(prefix=W.prefix("network")))

            server.disconnect()
            break

        try:
            server.client.receive(data)
        except (RemoteTransportError, RemoteProtocolError) as e:
            server.error(str(e))
            server.disconnect()
            break

        response = server.client.next_response()

        # Check if we need to send some data back
        data_to_send = server.client.data_to_send()

        if data_to_send:
            server.send(data_to_send)

        if response:
            server.handle_response(response)
            break

    return W.WEECHAT_RC_OK


def finalize_connection(server):
    hook = W.hook_fd(
        server.socket.fileno(),
        1,
        0,
        0,
        "receive_cb",
        server.name
    )

    server.fd_hook = hook
    server.connected = True
    server.connecting = False
    server.reconnect_delay = 0

    negotiated_protocol = (server.socket.selected_alpn_protocol() or
            server.socket.selected_npn_protocol())

    if negotiated_protocol == "h2":
        server.transport_type = TransportType.HTTP2
    else:
        server.transport_type = TransportType.HTTP

    data = server.client.connect(server.transport_type)
    server.send(data)

    server.login_info()


@utf8_decode
def sso_login_cb(server_name, command, return_code, out, err):
    try:
        server = SERVERS[server_name]
    except KeyError:
        message = (
            "{}{}: SSO callback ran, but no server for it was found.").format(
                W.prefix("error"), SCRIPT_NAME)
        W.prnt("", message)

    if return_code == W.WEECHAT_HOOK_PROCESS_ERROR:
        server.error("Error while running the matrix_sso_helper. Please "
                     "make sure that the helper script is executable and can "
                     "be found in your PATH.")
        server.sso_hook = None
        server.disconnect()
        return W.WEECHAT_RC_OK

    # The child process exited mark the hook as done.
    if return_code == 0:
        server.sso_hook = None

    if err != "":
        W.prnt("", "stderr: %s" % err)

    if out == "":
        return W.WEECHAT_RC_OK

    try:
        ret = json.loads(out)
        msgtype = ret.get("type")

        if msgtype == "redirectUrl":
            redirect_url = "http://{}:{}".format(ret["host"], ret["port"])

            login_url = (
                "{}/_matrix/client/r0/login/sso/redirect?redirectUrl={}"
            ).format(server.homeserver.geturl(), redirect_url)

            server.info_highlight(
                "The server requested a single sign-on, please open "
                "this URL in your browser. Note that the "
                "browser needs to run on the same host as Weechat.")
            server.info_highlight(login_url)

            message = {
                "server": server.name,
                "url": login_url
            }
            W.hook_hsignal_send("matrix_sso_login", message)

        elif msgtype == "token":
            token = ret["loginToken"]
            server.login(token=token)

        elif msgtype == "error":
            server.error("Error in the SSO helper {}".format(ret["message"]))

        else:
            server.error("Unknown SSO login message received from child "
                         "process.")

    except JSONDecodeError:
        server.error(
            "Error decoding SSO login message from child process: {}".format(
                out
        ))

    return W.WEECHAT_RC_OK


@utf8_decode
def connect_cb(data, status, gnutls_rc, sock, error, ip_address):
    # pylint: disable=too-many-arguments,too-many-branches
    status_value = int(status)  # type: int
    server = SERVERS[data]

    if status_value == W.WEECHAT_HOOK_CONNECT_OK:
        file_descriptor = int(sock)  # type: int
        server.numeric_address = ip_address
        server_buffer_set_title(server)

        wrap_socket(server, file_descriptor)

        return W.WEECHAT_RC_OK

    elif status_value == W.WEECHAT_HOOK_CONNECT_ADDRESS_NOT_FOUND:
        server.error('{address} not found'.format(address=ip_address))

    elif status_value == W.WEECHAT_HOOK_CONNECT_IP_ADDRESS_NOT_FOUND:
        server.error('IP address not found')

    elif status_value == W.WEECHAT_HOOK_CONNECT_CONNECTION_REFUSED:
        server.error('Connection refused')

    elif status_value == W.WEECHAT_HOOK_CONNECT_PROXY_ERROR:
        server.error('Proxy fails to establish connection to server')

    elif status_value == W.WEECHAT_HOOK_CONNECT_LOCAL_HOSTNAME_ERROR:
        server.error('Unable to set local hostname')

    elif status_value == W.WEECHAT_HOOK_CONNECT_GNUTLS_INIT_ERROR:
        server.error('TLS init error')

    elif status_value == W.WEECHAT_HOOK_CONNECT_GNUTLS_HANDSHAKE_ERROR:
        server.error('TLS Handshake failed')

    elif status_value == W.WEECHAT_HOOK_CONNECT_MEMORY_ERROR:
        server.error('Not enough memory')

    elif status_value == W.WEECHAT_HOOK_CONNECT_TIMEOUT:
        server.error('Timeout')

    elif status_value == W.WEECHAT_HOOK_CONNECT_SOCKET_ERROR:
        server.error('Unable to create socket')
    else:
        server.error('Unexpected error: {status}'.format(status=status_value))

    server.disconnect(reconnect=True)
    return W.WEECHAT_RC_OK


@utf8_decode
def room_close_cb(data, buffer):
    W.prnt("",
           "Buffer '%s' will be closed!" % W.buffer_get_string(buffer, "name"))
    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_unload_cb():
    for server in SERVERS.values():
        server.config.free()

    G.CONFIG.free()

    return W.WEECHAT_RC_OK


def autoconnect(servers):
    for server in servers.values():
        if server.config.autoconnect:
            server.connect()


def debug_buffer_close_cb(data, buffer):
    G.CONFIG.debug_buffer = ""
    return W.WEECHAT_RC_OK


def server_buffer_cb(server_name, buffer, input_data):
    message = ("{}{}: this buffer is not a room buffer!").format(
            W.prefix("error"), SCRIPT_NAME)
    W.prnt(buffer, message)
    return W.WEECHAT_RC_OK


class WeechatHandler(StreamHandler):
    def __init__(self, level=logbook.NOTSET, format_string=None, filter=None,
                 bubble=False):
        StreamHandler.__init__(
            self,
            object(),
            level,
            format_string,
            None,
            filter,
            bubble
        )

    def write(self, item):
        buf = ""

        if G.CONFIG.network.debug_buffer:
            if not G.CONFIG.debug_buffer:
                G.CONFIG.debug_buffer = W.buffer_new(
                    "Matrix Debug", "", "", "debug_buffer_close_cb", "")

            buf = G.CONFIG.debug_buffer

        W.prnt(buf, item)


def buffer_switch_cb(_, _signal, buffer_ptr):
    """Do some buffer operations when we switch buffers.

    This function is called every time we switch a buffer. The pointer of
    the new buffer is given to us by weechat.

    If it is one of our room buffers we check if the members for the room
    aren't fetched and fetch them now if they aren't.

    Read receipts are send out from here as well.
    """
    for server in SERVERS.values():
        if buffer_ptr == server.server_buffer:
            return W.WEECHAT_RC_OK

        if buffer_ptr not in server.buffers.values():
            continue

        room_buffer = server.find_room_from_ptr(buffer_ptr)
        if not room_buffer:
            continue

        last_event_id = room_buffer.last_event_id

        if room_buffer.should_send_read_marker:
            # A buffer may not have any events, in that case no event id is
            # here returned
            if last_event_id:
                server.room_send_read_marker(
                    room_buffer.room.room_id, last_event_id)
                room_buffer.last_read_event = last_event_id

        if not room_buffer.members_fetched:
            room_id = room_buffer.room.room_id
            server.get_joined_members(room_id)

        # The buffer is empty and we are seeing it for the first time.
        # Let us fetch some messages from the room history so it doesn't feel so
        # empty.
        if room_buffer.first_view and room_buffer.weechat_buffer.num_lines < 10:
            # TODO we may want to fetch 10 - num_lines messages here for
            # consistency reasons.
            server.room_get_messages(room_buffer.room.room_id)

        break

    return W.WEECHAT_RC_OK


def typing_notification_cb(data, signal, buffer_ptr):
    """Send out typing notifications if the user is typing.

    This function is called every time the input text is changed.
    It checks if we are on a buffer we own, and if we are sends out a typing
    notification if the room is configured to send them out.
    """
    for server in SERVERS.values():
        room_buffer = server.find_room_from_ptr(buffer_ptr)
        if room_buffer:
            server.room_send_typing_notice(room_buffer)
            return W.WEECHAT_RC_OK

        if buffer_ptr == server.server_buffer:
            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


def buffer_command_cb(data, _, command):
    """Override the buffer command to allow switching buffers by short name."""
    command = command[7:].strip()

    buffer_subcommands = ["list", "add", "clear", "move", "swap", "cycle",
                          "merge", "unmerge", "hide", "unhide", "renumber",
                          "close", "notify", "localvar", "set", "get"]

    if not command:
        return W.WEECHAT_RC_OK

    command_words = command.split()

    if len(command_words) >= 1:
        if command_words[0] in buffer_subcommands:
            return W.WEECHAT_RC_OK

        elif command_words[0].startswith("*"):
            return W.WEECHAT_RC_OK

        try:
            int(command_words[0])
            return W.WEECHAT_RC_OK
        except ValueError:
            pass

    room_buffers = []

    for server in SERVERS.values():
        room_buffers.extend(server.room_buffers.values())

    sorted_buffers = sorted(
        room_buffers,
        key=lambda b: b.weechat_buffer.number
    )

    for room_buffer in sorted_buffers:
        buffer = room_buffer.weechat_buffer

        if command in buffer.short_name:
            displayed = W.current_buffer() == buffer._ptr

            if displayed:
                continue

            W.buffer_set(buffer._ptr, 'display', '1')
            return W.WEECHAT_RC_OK_EAT

    return W.WEECHAT_RC_OK


if __name__ == "__main__":
    if W.register(WEECHAT_SCRIPT_NAME, WEECHAT_SCRIPT_AUTHOR,
                  WEECHAT_SCRIPT_VERSION, WEECHAT_SCRIPT_LICENSE,
                  WEECHAT_SCRIPT_DESCRIPTION, 'matrix_unload_cb', ''):

        if not W.mkdir_home("matrix", 0o700):
            message = ("{prefix}matrix: Error creating session "
                       "directory").format(prefix=W.prefix("error"))
            W.prnt("", message)

        handler = WeechatHandler()
        handler.format_string = "{record.channel}: {record.message}"
        handler.push_application()

        # TODO if this fails we should abort and unload the script.
        G.CONFIG = MatrixConfig()
        G.CONFIG.read()

        hook_commands()
        hook_key_bindings()
        init_bar_items()
        init_completion()

        W.hook_command_run("/buffer", "buffer_command_cb", "")
        W.hook_signal("buffer_switch", "buffer_switch_cb", "")
        W.hook_signal("input_text_changed", "typing_notification_cb", "")

        if not SERVERS:
            create_default_server(G.CONFIG)

        autoconnect(SERVERS)



# -*- coding: utf-8 -*-

# Copyright © 2008 Nicholas Marriott <nicholas.marriott@gmail.com>
# Copyright © 2016 Avi Halachmi <avihpit@yahoo.com>
# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
# Copyright © 2018, 2019 Denis Kasak <dkasak@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals

import html
import re
import textwrap

# pylint: disable=redefined-builtin
from builtins import str
from collections import namedtuple
from typing import Dict, List, Optional, Union

import webcolors
from pygments import highlight
from pygments.formatter import Formatter, get_style_by_name
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from . import globals as G
from .globals import W
from .utils import (string_strikethrough,
                    string_color_and_reset,
                    color_pair,
                    text_block,
                    colored_text_block)

try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser


class FormattedString:
    __slots__ = ("text", "attributes")

    def __init__(self, text, attributes):
        self.attributes = DEFAULT_ATTRIBUTES.copy()
        self.attributes.update(attributes)
        self.text = text


class Formatted(object):
    def __init__(self, substrings):
        # type: (List[FormattedString]) -> None
        self.substrings = substrings

    def textwrapper(self, width, colors):
        return textwrap.TextWrapper(
            width=width,
            initial_indent="{}> ".format(W.color(colors)),
            subsequent_indent="{}> ".format(W.color(colors)),
        )

    def is_formatted(self):
        # type: (Formatted) -> bool
        for string in self.substrings:
            if string.attributes != DEFAULT_ATTRIBUTES:
                return True
        return False

    # TODO reverse video
    @classmethod
    def from_input_line(cls, line):
        # type: (str) -> Formatted
        """Parses the weechat input line and produces formatted strings that
        can be later converted to HTML or to a string for weechat's print
        functions
        """
        text = ""  # type: str
        substrings = []  # type: List[FormattedString]
        attributes = DEFAULT_ATTRIBUTES.copy()

        # If this is false, only IRC formatting characters will be parsed.
        do_markdown = G.CONFIG.look.markdown_input

        # Disallow backticks in URLs so that code blocks are unaffected by the
        # URL handling
        url_regex = r"\b[a-z]+://[^\s`]+"

        # Escaped things are not markdown delimiters, so substitute them away
        # when (quickly) looking for the last delimiters in the line.
        # Additionally, URLs are ignored for the purposes of markdown
        # delimiters.
        # Note that the replacement needs to be the same length as the original
        # for the indices to be correct.
        escaped_masked = re.sub(
            r"\\[\\*_`]|(?:" + url_regex + ")",
            lambda m: "a" * len(m.group(0)),
            line
        )

        def last_match_index(regex, offset_in_match):
            matches = list(re.finditer(regex, escaped_masked))
            return matches[-1].span()[0] + offset_in_match if matches else -1

        # 'needs_word': whether the wrapper must surround words, for example
        #   '*italic*' and not '* not-italic *'.
        # 'validate': whether it can occur within the current attributes
        wrappers = {
            "**": {
                "key": "bold",
                "last_index": last_match_index(r"\S\*\*", 1),
                "needs_word": True,
                "validate": lambda attrs: not attrs["code"],
            },
            "*": {
                "key": "italic",
                "last_index": last_match_index(r"\S\*($|[^*])", 1),
                "needs_word": True,
                "validate": lambda attrs: not attrs["code"],
            },
            "_": {
                "key": "italic",
                "last_index": last_match_index(r"\S_", 1),
                "needs_word": True,
                "validate": lambda attrs: not attrs["code"],
            },
            "`": {
                "key": "code",
                "last_index": last_match_index(r"`", 0),
                "needs_word": False,
                "validate": lambda attrs: True,
            }
        }
        wrapper_init_chars = set(k[0] for k in wrappers.keys())
        wrapper_max_len = max(len(k) for k in wrappers.keys())

        irc_toggles = {
            "\x02": "bold",
            "\x1D": "italic",
            "\x1F": "underline",
        }

        # Characters that consume a prefixed backslash
        escapable_chars = wrapper_init_chars.copy()
        escapable_chars.add("\\")

        # Collect URL spans
        url_spans = [m.span() for m in re.finditer(url_regex, line)]
        url_spans.reverse()  # we'll be popping from the end

        # Whether we are currently in a URL
        in_url = False

        i = 0
        while i < len(line):
            # Update the 'in_url' flag. The first condition is not a while loop
            # because URLs must contain '://', ensuring that we will not skip 2
            # URLs in one iteration.
            if url_spans and i >= url_spans[-1][1]:
                in_url = False
                url_spans.pop()
            if url_spans and i >= url_spans[-1][0]:
                in_url = True

            # Markdown escape
            if do_markdown and \
                    i + 1 < len(line) and line[i] == "\\" \
                    and (line[i + 1] in escapable_chars
                            if not attributes["code"]
                            else line[i + 1] == "`") \
                    and not in_url:
                text += line[i + 1]
                i = i + 2

            # IRC bold/italic/underline
            elif line[i] in irc_toggles and not attributes["code"]:
                if text:
                    substrings.append(FormattedString(text, attributes.copy()))
                text = ""
                key = irc_toggles[line[i]]
                attributes[key] = not attributes[key]
                i = i + 1

            # IRC reset
            elif line[i] == "\x0F" and not attributes["code"]:
                if text:
                    substrings.append(FormattedString(text, attributes.copy()))
                text = ""
                # Reset all the attributes
                attributes = DEFAULT_ATTRIBUTES.copy()
                i = i + 1

            # IRC color
            elif line[i] == "\x03" and not attributes["code"]:
                if text:
                    substrings.append(FormattedString(text, attributes.copy()))
                text = ""
                i = i + 1

                # check if it's a valid color, add it to the attributes
                if line[i].isdigit():
                    color_string = line[i]
                    i = i + 1

                    if line[i].isdigit():
                        if color_string == "0":
                            color_string = line[i]
                        else:
                            color_string = color_string + line[i]
                        i = i + 1

                    attributes["fgcolor"] = color_line_to_weechat(color_string)
                else:
                    attributes["fgcolor"] = None

                # check if we have a background color
                if line[i] == "," and line[i + 1].isdigit():
                    color_string = line[i + 1]
                    i = i + 2

                    if line[i].isdigit():
                        if color_string == "0":
                            color_string = line[i]
                        else:
                            color_string = color_string + line[i]
                        i = i + 1

                    attributes["bgcolor"] = color_line_to_weechat(color_string)
                else:
                    attributes["bgcolor"] = None

            # Markdown wrapper (emphasis/bold/code)
            elif do_markdown and line[i] in wrapper_init_chars and not in_url:
                for l in range(wrapper_max_len, 0, -1):
                    if i + l <= len(line) and line[i : i + l] in wrappers:
                        descriptor = wrappers[line[i : i + l]]

                        if not descriptor["validate"](attributes):
                            continue

                        if attributes[descriptor["key"]]:
                            # needs_word wrappers can only be turned off if
                            # preceded by non-whitespace
                            if (i >= 1 and not line[i - 1].isspace()) \
                                    or not descriptor["needs_word"]:
                                if text:
                                    # strip leading and trailing spaces and
                                    # compress consecutive spaces in inline
                                    # code blocks
                                    if descriptor["key"] == "code":
                                        text = re.sub(r"\s+", " ", text.strip())
                                    substrings.append(
                                        FormattedString(text, attributes.copy()))
                                text = ""
                                attributes[descriptor["key"]] = False
                                i = i + l
                            else:
                                text = text + line[i : i + l]
                                i = i + l

                        # Must have a chance of closing this, and needs_word
                        # wrappers must be followed by non-whitespace
                        elif descriptor["last_index"] >= i + l and \
                                (not line[i + l].isspace() or \
                                    not descriptor["needs_word"]):
                            if text:
                                substrings.append(
                                    FormattedString(text, attributes.copy()))
                            text = ""
                            attributes[descriptor["key"]] = True
                            i = i + l

                        else:
                            text = text + line[i : i + l]
                            i = i + l

                        break

                else:
                    # No wrapper matched here (NOTE: cannot happen since all
                    # wrapper prefixes are also wrappers, but for completeness'
                    # sake)
                    text = text + line[i]
                    i = i + 1

            # Normal text
            else:
                text = text + line[i]
                i = i + 1

        if text:
            substrings.append(FormattedString(text, attributes))

        return cls(substrings)

    @classmethod
    def from_html(cls, html):
        # type: (str) -> Formatted
        parser = MatrixHtmlParser()
        parser.feed(html)
        return cls(parser.get_substrings())

    def to_html(self):
        def add_attribute(string, name, value):
            if name == "bold" and value:
                return "{bold_on}{text}{bold_off}".format(
                    bold_on="<strong>", text=string, bold_off="</strong>"
                )
            if name == "italic" and value:
                return "{italic_on}{text}{italic_off}".format(
                    italic_on="<em>", text=string, italic_off="</em>"
                )
            if name == "underline" and value:
                return "{underline_on}{text}{underline_off}".format(
                    underline_on="<u>", text=string, underline_off="</u>"
                )
            if name == "strikethrough" and value:
                return "{strike_on}{text}{strike_off}".format(
                    strike_on="<del>", text=string, strike_off="</del>"
                )
            if name == "quote" and value:
                return "{quote_on}{text}{quote_off}".format(
                    quote_on="<blockquote>",
                    text=string,
                    quote_off="</blockquote>",
                )
            if name == "code" and value:
                return "{code_on}{text}{code_off}".format(
                    code_on="<code>", text=string, code_off="</code>"
                )

            return string

        def add_color(string, fgcolor, bgcolor):
            fgcolor_string = ""
            bgcolor_string = ""

            if fgcolor:
                fgcolor_string = " data-mx-color={}".format(
                    color_weechat_to_html(fgcolor)
                )

            if bgcolor:
                bgcolor_string = " data-mx-bg-color={}".format(
                    color_weechat_to_html(bgcolor)
                )

            return "{color_on}{text}{color_off}".format(
                color_on="<font{fg}{bg}>".format(
                    fg=fgcolor_string,
                    bg=bgcolor_string
                ),
                text=string,
                color_off="</font>",
            )

        def format_string(formatted_string):
            text = formatted_string.text
            attributes = formatted_string.attributes.copy()

            # Escape HTML tag characters
            text = text.replace("&", "&amp;") \
                       .replace("<", "&lt;") \
                       .replace(">", "&gt;")

            if attributes["code"]:
                if attributes["preformatted"]:
                    # XXX: This can't really happen since there's no way of
                    # creating preformatted code blocks in weechat (because
                    # there is not multiline input), but I'm creating this
                    # branch as a note that it should be handled once we do
                    # implement them.
                    pass
                else:
                    text = add_attribute(text, "code", True)
                    attributes.pop("code")

            if attributes["fgcolor"] or attributes["bgcolor"]:
                text = add_color(
                    text,
                    attributes["fgcolor"],
                    attributes["bgcolor"]
                )

                if attributes["fgcolor"]:
                    attributes.pop("fgcolor")

                if attributes["bgcolor"]:
                    attributes.pop("bgcolor")

            for key, value in attributes.items():
                text = add_attribute(text, key, value)

            return text

        html_string = map(format_string, self.substrings)
        return "".join(html_string)

    # TODO do we want at least some formatting using unicode
    # (strikethrough, quotes)?
    def to_plain(self):
        # type: () -> str
        def strip_atribute(string, _, __):
            return string

        def format_string(formatted_string):
            text = formatted_string.text
            attributes = formatted_string.attributes

            for key, value in attributes.items():
                text = strip_atribute(text, key, value)
            return text

        plain_string = map(format_string, self.substrings)
        return "".join(plain_string)

    def to_weechat(self):
        def add_attribute(string, name, value, attributes):
            if not value:
                return string
            elif name == "bold":
                return "{bold_on}{text}{bold_off}".format(
                    bold_on=W.color("bold"),
                    text=string,
                    bold_off=W.color("-bold"),
                )
            elif name == "italic":
                return "{italic_on}{text}{italic_off}".format(
                    italic_on=W.color("italic"),
                    text=string,
                    italic_off=W.color("-italic"),
                )
            elif name == "underline":
                return "{underline_on}{text}{underline_off}".format(
                    underline_on=W.color("underline"),
                    text=string,
                    underline_off=W.color("-underline"),
                )
            elif name == "strikethrough":
                return string_strikethrough(string)
            elif name == "quote":
                quote_pair = color_pair(G.CONFIG.color.quote_fg,
                                        G.CONFIG.color.quote_bg)

                # Remove leading and trailing newlines; Riot sends an extra
                # quoted "\n" when a user quotes a message.
                string = string.strip("\n")
                if len(string) == 0:
                    return string

                if G.CONFIG.look.quote_wrap >= 0:
                    wrapper = self.textwrapper(G.CONFIG.look.quote_wrap, quote_pair)
                    return wrapper.fill(W.string_remove_color(string, ""))
                else:
                    # Don't wrap, just add quote markers to all lines
                    return "{color_on}{text}{color_off}".format(
                        color_on=W.color(quote_pair),
                        text="> " + W.string_remove_color(string.replace("\n", "\n> "), ""),
                        color_off=W.color("resetcolor")
                    )
            elif name == "code":
                code_color_pair = color_pair(
                    G.CONFIG.color.untagged_code_fg,
                    G.CONFIG.color.untagged_code_bg
                )

                margin = G.CONFIG.look.code_block_margin

                if attributes["preformatted"]:
                    # code block

                    try:
                        lexer = get_lexer_by_name(value)
                    except ClassNotFound:
                        if G.CONFIG.look.code_blocks:
                            return colored_text_block(
                                string,
                                margin=margin,
                                color_pair=code_color_pair)
                        else:
                            return string_color_and_reset(string,
                                                          code_color_pair)

                    try:
                        style = get_style_by_name(G.CONFIG.look.pygments_style)
                    except ClassNotFound:
                        style = "native"

                    if G.CONFIG.look.code_blocks:
                        code_block = text_block(string, margin=margin)
                    else:
                        code_block = string

                    # highlight adds a newline to the end of the string, remove
                    # it from the output
                    highlighted_code = highlight(
                        code_block,
                        lexer,
                        WeechatFormatter(style=style)
                    ).rstrip()

                    return highlighted_code
                else:
                    return string_color_and_reset(string, code_color_pair)
            elif name == "fgcolor":
                return "{color_on}{text}{color_off}".format(
                    color_on=W.color(value),
                    text=string,
                    color_off=W.color("resetcolor"),
                )
            elif name == "bgcolor":
                return "{color_on}{text}{color_off}".format(
                    color_on=W.color("," + value),
                    text=string,
                    color_off=W.color("resetcolor"),
                )
            else:
                return string

        def format_string(formatted_string):
            text = formatted_string.text
            attributes = formatted_string.attributes

            # We need to handle strikethrough first, since doing
            # a strikethrough followed by other attributes succeeds in the
            # terminal, but doing it the other way around results in garbage.
            if "strikethrough" in attributes:
                text = add_attribute(
                    text,
                    "strikethrough",
                    attributes["strikethrough"],
                    attributes
                )
                attributes.pop("strikethrough")

            def indent(text, prefix):
                return prefix + text.replace("\n", "\n{}".format(prefix))

            for key, value in attributes.items():
                if not value:
                    continue

                # Don't use textwrap to quote the code
                if key == "quote" and attributes["code"]:
                    continue

                # Reflow inline code blocks
                if key == "code" and not attributes["preformatted"]:
                    text = text.strip().replace('\n', ' ')

                text = add_attribute(text, key, value, attributes)

                # If we're quoted code add quotation marks now.
                if key == "code" and attributes["quote"]:
                    fg = G.CONFIG.color.quote_fg
                    bg = G.CONFIG.color.quote_bg
                    text = indent(
                        text,
                        string_color_and_reset(">", color_pair(fg, bg)) + " ",
                    )

            # If we're code don't remove multiple newlines blindly
            if attributes["code"]:
                return text
            return re.sub(r"\n+", "\n", text)

        weechat_strings = map(format_string, self.substrings)

        # Remove duplicate \n elements from the list
        strings = []
        for string in weechat_strings:
            if len(strings) == 0 or string != "\n" or string != strings[-1]:
                strings.append(string)

        return "".join(strings).strip()


DEFAULT_ATTRIBUTES = {
    "bold": False,
    "italic": False,
    "underline": False,
    "strikethrough": False,
    "preformatted": False,
    "quote": False,
    "code": None,
    "fgcolor": None,
    "bgcolor": None,
}  # type: Dict[str, Union[bool, Optional[str]]]


class MatrixHtmlParser(HTMLParser):
    # TODO bullets
    def __init__(self):
        HTMLParser.__init__(self)
        self.text = ""  # type: str
        self.substrings = []  # type: List[FormattedString]
        self.attributes = DEFAULT_ATTRIBUTES.copy()

    def unescape(self, text):
        """Shim to unescape HTML in both Python 2 and 3.

        The instance method was deprecated in Python 3 and html.unescape
        doesn't exist in Python 2 so this is needed.
        """
        try:
            return html.unescape(text)
        except AttributeError:
            return HTMLParser.unescape(self, text)

    def add_substring(self, text, attrs):
        fmt_string = FormattedString(text, attrs)
        self.substrings.append(fmt_string)

    def _toggle_attribute(self, attribute):
        if self.text:
            self.add_substring(self.text, self.attributes.copy())
        self.text = ""
        self.attributes[attribute] = not self.attributes[attribute]

    def handle_starttag(self, tag, attrs):
        if tag == "strong":
            self._toggle_attribute("bold")
        elif tag == "em":
            self._toggle_attribute("italic")
        elif tag == "u":
            self._toggle_attribute("underline")
        elif tag == "del":
            self._toggle_attribute("strikethrough")
        elif tag == "blockquote":
            self._toggle_attribute("quote")
        elif tag == "pre":
            self._toggle_attribute("preformatted")
        elif tag == "code":
            lang = None

            for key, value in attrs:
                if key == "class":
                    if value.startswith("language-"):
                        lang = value.split("-", 1)[1]

            lang = lang or "unknown"

            if self.text:
                self.add_substring(self.text, self.attributes.copy())
            self.text = ""
            self.attributes["code"] = lang
        elif tag == "p":
            if self.text:
                self.add_substring(self.text, self.attributes.copy())
            self.text = "\n"
            self.add_substring(self.text, DEFAULT_ATTRIBUTES.copy())
            self.text = ""
        elif tag == "br":
            if self.text:
                self.add_substring(self.text, self.attributes.copy())
            self.text = "\n"
            self.add_substring(self.text, DEFAULT_ATTRIBUTES.copy())
            self.text = ""
        elif tag == "font":
            for key, value in attrs:
                if key in ["data-mx-color", "color"]:
                    color = color_html_to_weechat(value)

                    if not color:
                        continue

                    if self.text:
                        self.add_substring(self.text, self.attributes.copy())
                    self.text = ""
                    self.attributes["fgcolor"] = color

                elif key in ["data-mx-bg-color"]:
                    color = color_html_to_weechat(value)
                    if not color:
                        continue

                    if self.text:
                        self.add_substring(self.text, self.attributes.copy())
                    self.text = ""
                    self.attributes["bgcolor"] = color

        else:
            pass

    def handle_endtag(self, tag):
        if tag == "strong":
            self._toggle_attribute("bold")
        elif tag == "em":
            self._toggle_attribute("italic")
        elif tag == "u":
            self._toggle_attribute("underline")
        elif tag == "del":
            self._toggle_attribute("strikethrough")
        elif tag == "pre":
            self._toggle_attribute("preformatted")
        elif tag == "code":
            if self.text:
                self.add_substring(self.text, self.attributes.copy())
            self.text = ""
            self.attributes["code"] = None
        elif tag == "blockquote":
            self._toggle_attribute("quote")
            self.text = "\n"
            self.add_substring(self.text, DEFAULT_ATTRIBUTES.copy())
            self.text = ""
        elif tag == "font":
            if self.text:
                self.add_substring(self.text, self.attributes.copy())
            self.text = ""
            self.attributes["fgcolor"] = None
        else:
            pass

    def handle_data(self, data):
        self.text += data

    def handle_entityref(self, name):
        self.text += self.unescape("&{};".format(name))

    def handle_charref(self, name):
        self.text += self.unescape("&#{};".format(name))

    def get_substrings(self):
        if self.text:
            self.add_substring(self.text, self.attributes.copy())

        return self.substrings


def color_line_to_weechat(color_string):
    # type: (str) -> str
    line_colors = {
        "0": "white",
        "1": "black",
        "2": "blue",
        "3": "green",
        "4": "lightred",
        "5": "red",
        "6": "magenta",
        "7": "brown",
        "8": "yellow",
        "9": "lightgreen",
        "10": "cyan",
        "11": "lightcyan",
        "12": "lightblue",
        "13": "lightmagenta",
        "14": "darkgray",
        "15": "gray",
        "16": "52",
        "17": "94",
        "18": "100",
        "19": "58",
        "20": "22",
        "21": "29",
        "22": "23",
        "23": "24",
        "24": "17",
        "25": "54",
        "26": "53",
        "27": "89",
        "28": "88",
        "29": "130",
        "30": "142",
        "31": "64",
        "32": "28",
        "33": "35",
        "34": "30",
        "35": "25",
        "36": "18",
        "37": "91",
        "38": "90",
        "39": "125",
        "40": "124",
        "41": "166",
        "42": "184",
        "43": "106",
        "44": "34",
        "45": "49",
        "46": "37",
        "47": "33",
        "48": "19",
        "49": "129",
        "50": "127",
        "51": "161",
        "52": "196",
        "53": "208",
        "54": "226",
        "55": "154",
        "56": "46",
        "57": "86",
        "58": "51",
        "59": "75",
        "60": "21",
        "61": "171",
        "62": "201",
        "63": "198",
        "64": "203",
        "65": "215",
        "66": "227",
        "67": "191",
        "68": "83",
        "69": "122",
        "70": "87",
        "71": "111",
        "72": "63",
        "73": "177",
        "74": "207",
        "75": "205",
        "76": "217",
        "77": "223",
        "78": "229",
        "79": "193",
        "80": "157",
        "81": "158",
        "82": "159",
        "83": "153",
        "84": "147",
        "85": "183",
        "86": "219",
        "87": "212",
        "88": "16",
        "89": "233",
        "90": "235",
        "91": "237",
        "92": "239",
        "93": "241",
        "94": "244",
        "95": "247",
        "96": "250",
        "97": "254",
        "98": "231",
        "99": "default",
    }

    assert color_string in line_colors

    return line_colors[color_string]


# The functions color_dist_sq(), color_to_6cube(), and color_find_rgb
# are python ports of the same named functions from the tmux
# source, they are under the copyright of Nicholas Marriott, and Avi Halachmi
# under the ISC license.
# More info: https://github.com/tmux/tmux/blob/master/colour.c


def color_dist_sq(R, G, B, r, g, b):
    # pylint: disable=invalid-name,too-many-arguments
    # type: (int, int, int, int, int, int) -> int
    return (R - r) * (R - r) + (G - g) * (G - g) + (B - b) * (B - b)


def color_to_6cube(v):
    # pylint: disable=invalid-name
    # type: (int) -> int
    if v < 48:
        return 0
    if v < 114:
        return 1
    return (v - 35) // 40


def color_find_rgb(r, g, b):
    # type: (int, int, int) -> int
    """Convert an RGB triplet to the xterm(1) 256 color palette.

       xterm provides a 6x6x6 color cube (16 - 231) and 24 greys (232 - 255).
       We map our RGB color to the closest in the cube, also work out the
       closest grey, and use the nearest of the two.

       Note that the xterm has much lower resolution for darker colors (they
       are not evenly spread out), so our 6 levels are not evenly spread: 0x0,
       0x5f (95), 0x87 (135), 0xaf (175), 0xd7 (215) and 0xff (255). Greys are
       more evenly spread (8, 18, 28 ... 238).
    """
    # pylint: disable=invalid-name
    q2c = [0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff]

    # Map RGB to 6x6x6 cube.
    qr = color_to_6cube(r)
    qg = color_to_6cube(g)
    qb = color_to_6cube(b)

    cr = q2c[qr]
    cg = q2c[qg]
    cb = q2c[qb]

    # If we have hit the color exactly, return early.
    if cr == r and cg == g and cb == b:
        return 16 + (36 * qr) + (6 * qg) + qb

    # Work out the closest grey (average of RGB).
    grey_avg = (r + g + b) // 3

    if grey_avg > 238:
        grey_idx = 23
    else:
        grey_idx = (grey_avg - 3) // 10

    grey = 8 + (10 * grey_idx)

    # Is grey or 6x6x6 color closest?
    d = color_dist_sq(cr, cg, cb, r, g, b)

    if color_dist_sq(grey, grey, grey, r, g, b) < d:
        idx = 232 + grey_idx
    else:
        idx = 16 + (36 * qr) + (6 * qg) + qb

    return idx


def color_html_to_weechat(color):
    # type: (str) -> str
    # yapf: disable
    weechat_basic_colors = {
        (0, 0, 0): "black",             # 0
        (128, 0, 0): "red",             # 1
        (0, 128, 0): "green",           # 2
        (128, 128, 0): "brown",         # 3
        (0, 0, 128): "blue",            # 4
        (128, 0, 128): "magenta",       # 5
        (0, 128, 128): "cyan",          # 6
        (192, 192, 192): "default",     # 7
        (128, 128, 128): "gray",        # 8
        (255, 0, 0): "lightred",        # 9
        (0, 255, 0): "lightgreen",      # 10
        (255, 255, 0): "yellow",        # 11
        (0, 0, 255): "lightblue",       # 12
        (255, 0, 255): "lightmagenta",  # 13
        (0, 255, 255): "lightcyan",     # 14
        (255, 255, 255): "white",       # 15
    }
    # yapf: enable

    try:
        rgb_color = webcolors.html5_parse_legacy_color(color)
    except ValueError:
        return ""

    if rgb_color in weechat_basic_colors:
        return weechat_basic_colors[rgb_color]

    return str(color_find_rgb(*rgb_color))


def color_weechat_to_html(color):
    # type: (str) -> str
    # yapf: disable
    weechat_basic_colors = {
        "black": "0",
        "red": "1",
        "green": "2",
        "brown": "3",
        "blue": "4",
        "magenta": "5",
        "cyan": "6",
        "default": "7",
        "gray": "8",
        "lightred": "9",
        "lightgreen": "10",
        "yellow": "11",
        "lightblue": "12",
        "lightmagenta": "13",
        "lightcyan": "14",
        "white": "15",
    }
    hex_colors = {
        "0":   "#000000",
        "1":   "#800000",
        "2":   "#008000",
        "3":   "#808000",
        "4":   "#000080",
        "5":   "#800080",
        "6":   "#008080",
        "7":   "#c0c0c0",
        "8":   "#808080",
        "9":   "#ff0000",
        "10":  "#00ff00",
        "11":  "#ffff00",
        "12":  "#0000ff",
        "13":  "#ff00ff",
        "14":  "#00ffff",
        "15":  "#ffffff",
        "16":  "#000000",
        "17":  "#00005f",
        "18":  "#000087",
        "19":  "#0000af",
        "20":  "#0000d7",
        "21":  "#0000ff",
        "22":  "#005f00",
        "23":  "#005f5f",
        "24":  "#005f87",
        "25":  "#005faf",
        "26":  "#005fd7",
        "27":  "#005fff",
        "28":  "#008700",
        "29":  "#00875f",
        "30":  "#008787",
        "31":  "#0087af",
        "32":  "#0087d7",
        "33":  "#0087ff",
        "34":  "#00af00",
        "35":  "#00af5f",
        "36":  "#00af87",
        "37":  "#00afaf",
        "38":  "#00afd7",
        "39":  "#00afff",
        "40":  "#00d700",
        "41":  "#00d75f",
        "42":  "#00d787",
        "43":  "#00d7af",
        "44":  "#00d7d7",
        "45":  "#00d7ff",
        "46":  "#00ff00",
        "47":  "#00ff5f",
        "48":  "#00ff87",
        "49":  "#00ffaf",
        "50":  "#00ffd7",
        "51":  "#00ffff",
        "52":  "#5f0000",
        "53":  "#5f005f",
        "54":  "#5f0087",
        "55":  "#5f00af",
        "56":  "#5f00d7",
        "57":  "#5f00ff",
        "58":  "#5f5f00",
        "59":  "#5f5f5f",
        "60":  "#5f5f87",
        "61":  "#5f5faf",
        "62":  "#5f5fd7",
        "63":  "#5f5fff",
        "64":  "#5f8700",
        "65":  "#5f875f",
        "66":  "#5f8787",
        "67":  "#5f87af",
        "68":  "#5f87d7",
        "69":  "#5f87ff",
        "70":  "#5faf00",
        "71":  "#5faf5f",
        "72":  "#5faf87",
        "73":  "#5fafaf",
        "74":  "#5fafd7",
        "75":  "#5fafff",
        "76":  "#5fd700",
        "77":  "#5fd75f",
        "78":  "#5fd787",
        "79":  "#5fd7af",
        "80":  "#5fd7d7",
        "81":  "#5fd7ff",
        "82":  "#5fff00",
        "83":  "#5fff5f",
        "84":  "#5fff87",
        "85":  "#5fffaf",
        "86":  "#5fffd7",
        "87":  "#5fffff",
        "88":  "#870000",
        "89":  "#87005f",
        "90":  "#870087",
        "91":  "#8700af",
        "92":  "#8700d7",
        "93":  "#8700ff",
        "94":  "#875f00",
        "95":  "#875f5f",
        "96":  "#875f87",
        "97":  "#875faf",
        "98":  "#875fd7",
        "99":  "#875fff",
        "100": "#878700",
        "101": "#87875f",
        "102": "#878787",
        "103": "#8787af",
        "104": "#8787d7",
        "105": "#8787ff",
        "106": "#87af00",
        "107": "#87af5f",
        "108": "#87af87",
        "109": "#87afaf",
        "110": "#87afd7",
        "111": "#87afff",
        "112": "#87d700",
        "113": "#87d75f",
        "114": "#87d787",
        "115": "#87d7af",
        "116": "#87d7d7",
        "117": "#87d7ff",
        "118": "#87ff00",
        "119": "#87ff5f",
        "120": "#87ff87",
        "121": "#87ffaf",
        "122": "#87ffd7",
        "123": "#87ffff",
        "124": "#af0000",
        "125": "#af005f",
        "126": "#af0087",
        "127": "#af00af",
        "128": "#af00d7",
        "129": "#af00ff",
        "130": "#af5f00",
        "131": "#af5f5f",
        "132": "#af5f87",
        "133": "#af5faf",
        "134": "#af5fd7",
        "135": "#af5fff",
        "136": "#af8700",
        "137": "#af875f",
        "138": "#af8787",
        "139": "#af87af",
        "140": "#af87d7",
        "141": "#af87ff",
        "142": "#afaf00",
        "143": "#afaf5f",
        "144": "#afaf87",
        "145": "#afafaf",
        "146": "#afafd7",
        "147": "#afafff",
        "148": "#afd700",
        "149": "#afd75f",
        "150": "#afd787",
        "151": "#afd7af",
        "152": "#afd7d7",
        "153": "#afd7ff",
        "154": "#afff00",
        "155": "#afff5f",
        "156": "#afff87",
        "157": "#afffaf",
        "158": "#afffd7",
        "159": "#afffff",
        "160": "#d70000",
        "161": "#d7005f",
        "162": "#d70087",
        "163": "#d700af",
        "164": "#d700d7",
        "165": "#d700ff",
        "166": "#d75f00",
        "167": "#d75f5f",
        "168": "#d75f87",
        "169": "#d75faf",
        "170": "#d75fd7",
        "171": "#d75fff",
        "172": "#d78700",
        "173": "#d7875f",
        "174": "#d78787",
        "175": "#d787af",
        "176": "#d787d7",
        "177": "#d787ff",
        "178": "#d7af00",
        "179": "#d7af5f",
        "180": "#d7af87",
        "181": "#d7afaf",
        "182": "#d7afd7",
        "183": "#d7afff",
        "184": "#d7d700",
        "185": "#d7d75f",
        "186": "#d7d787",
        "187": "#d7d7af",
        "188": "#d7d7d7",
        "189": "#d7d7ff",
        "190": "#d7ff00",
        "191": "#d7ff5f",
        "192": "#d7ff87",
        "193": "#d7ffaf",
        "194": "#d7ffd7",
        "195": "#d7ffff",
        "196": "#ff0000",
        "197": "#ff005f",
        "198": "#ff0087",
        "199": "#ff00af",
        "200": "#ff00d7",
        "201": "#ff00ff",
        "202": "#ff5f00",
        "203": "#ff5f5f",
        "204": "#ff5f87",
        "205": "#ff5faf",
        "206": "#ff5fd7",
        "207": "#ff5fff",
        "208": "#ff8700",
        "209": "#ff875f",
        "210": "#ff8787",
        "211": "#ff87af",
        "212": "#ff87d7",
        "213": "#ff87ff",
        "214": "#ffaf00",
        "215": "#ffaf5f",
        "216": "#ffaf87",
        "217": "#ffafaf",
        "218": "#ffafd7",
        "219": "#ffafff",
        "220": "#ffd700",
        "221": "#ffd75f",
        "222": "#ffd787",
        "223": "#ffd7af",
        "224": "#ffd7d7",
        "225": "#ffd7ff",
        "226": "#ffff00",
        "227": "#ffff5f",
        "228": "#ffff87",
        "229": "#ffffaf",
        "230": "#ffffd7",
        "231": "#ffffff",
        "232": "#080808",
        "233": "#121212",
        "234": "#1c1c1c",
        "235": "#262626",
        "236": "#303030",
        "237": "#3a3a3a",
        "238": "#444444",
        "239": "#4e4e4e",
        "240": "#585858",
        "241": "#626262",
        "242": "#6c6c6c",
        "243": "#767676",
        "244": "#808080",
        "245": "#8a8a8a",
        "246": "#949494",
        "247": "#9e9e9e",
        "248": "#a8a8a8",
        "249": "#b2b2b2",
        "250": "#bcbcbc",
        "251": "#c6c6c6",
        "252": "#d0d0d0",
        "253": "#dadada",
        "254": "#e4e4e4",
        "255": "#eeeeee"
    }

    # yapf: enable
    if color in weechat_basic_colors:
        return hex_colors[weechat_basic_colors[color]]
    return hex_colors[color]


class WeechatFormatter(Formatter):
    def __init__(self, **options):
        Formatter.__init__(self, **options)
        self.styles = {}

        for token, style in self.style:
            start = end = ""
            if style["color"]:
                start += "{}".format(
                    W.color(color_html_to_weechat(str(style["color"])))
                )
                end = "{}".format(W.color("resetcolor")) + end
            if style["bold"]:
                start += W.color("bold")
                end = W.color("-bold") + end
            if style["italic"]:
                start += W.color("italic")
                end = W.color("-italic") + end
            if style["underline"]:
                start += W.color("underline")
                end = W.color("-underline") + end
            self.styles[token] = (start, end)

    def format(self, tokensource, outfile):
        lastval = ""
        lasttype = None

        for ttype, value in tokensource:
            while ttype not in self.styles:
                ttype = ttype.parent

            if ttype == lasttype:
                lastval += value
            else:
                if lastval:
                    stylebegin, styleend = self.styles[lasttype]
                    outfile.write(stylebegin + lastval + styleend)
                # set lastval/lasttype to current values
                lastval = value
                lasttype = ttype

        if lastval:
            stylebegin, styleend = self.styles[lasttype]
            outfile.write(stylebegin + lastval + styleend)
